import copy
from enum import Enum

import matplotlib.pyplot as plt
import numpy as np
from numpy.polynomial.polynomial import polyfit
from scipy.stats import pearsonr

import deampy.in_out_functions as IO
import deampy.plots.histogram as Fig
import deampy.statistics as Stat
import deampy.support.misc_functions as F
from deampy.plots.plot_support import output_figure

# list of columns in the parameter csv file that are not considered a parameter
COLUMNS_TO_SKIP = ['ID', 'Seed', 'LnL', 'Probability', 'Message', 'Simulation Replication', 'Random Seed']
HISTOGRAM_FIG_SIZE = (4.2, 3.2)


class ColumnsPriorDistCSV(Enum):
    # columns of the csv file containing parameter prior distributions
    NAME = 0
    LB = 1
    UB = 2
    TITLE = 3
    MULTIPLIER = 4
    FORMAT = 5 # ',' to format as number, '%' to format as percentage, and '$' to format as currency
    DECI = 6


class ParamInfo:
    # class to store information about a parameter (name, estimate and confidence/uncertainty interval)
    def __init__(self, name, values=None, label=None, value_range=None):
        """
        :param name: (string) name of this parameter (mainly to use as a key in dictionaries)
        :param values: (list) of parameter values
        :param label: (string) label of this parameter to display on figures
        :param value_range: (tuple) range of this parameter to display on figures
        """
        self.name = name
        self.label = label
        self.values = np.array(values)
        self.range = value_range
        self.estimate = None
        self.interval = None


class ParameterAnalyzer:
    # class to analyze parameters (estimation and visualization)

    def __init__(self, csvfile_param_values=None, columns_to_be_deleted=()):
        """
        :param csvfile_param_values: (string) csv file where the parameter values are located
            assumes that the first row of this csv file contains the parameter names and each
            column contains the parameter values.
        :param columns_to_be_deleted: (list of string) list of column names to be deleted from analysis
        """

        # dictionary of parameter samples with parameter names as keys
        self.dictOfParamValues = {}

        if csvfile_param_values is not None:
            self.dictOfParamValues = IO.read_csv_cols_to_dictionary(file_name=csvfile_param_values,
                                                                    if_convert_float=True)
            for name in columns_to_be_deleted:
                del(self.dictOfParamValues[name])

    def resample_param_values(self, csvfile_param_values_and_weights,
                              n, weight_col, csvfile_resampled_params,
                              sample_by_weight=True, columns_to_be_deleted=(), seed=0):
        """
        :param csvfile_param_values_and_weights: (string) csv file where the values of parameters
            along with their weights are provided.
            It assumes that
                1) the first row contains the name of all parameters
                2) rows contains the weight and the parameter values
                3) rows are in decreasing order of parameter weights.
        :param n: (int) number of parameter values to resamples.
        :param weight_col: (int) index of the columns where the weights of parameter values are located.
        :param csvfile_resampled_params: (string) csvfile where the resampled parameter values will be stored.
            The first row will be the names of parameters.
        :param sample_by_weight: (bool) set to true to sample parameters by weight.
            If set to False, the first n parameters will be selected.
        :param columns_to_be_deleted: (list of string) list of column names to be deleted from analysis
        :param seed: (int) seed of the random number generator to resample parameters
        """

        # read parameter weights and values
        rows_of_weights_and_parameter_values = IO.read_csv_rows(
            file_name=csvfile_param_values_and_weights,
            if_ignore_first_row=False,
            if_convert_float=True)

        # store header
        rows_of_selected_param_values = [rows_of_weights_and_parameter_values[0]]

        if not sample_by_weight:
            # choose the first n parameter values
            for i, row in enumerate(rows_of_weights_and_parameter_values):
                if i > n: # first rwo is the header
                    break
                elif i > 0 and row[weight_col] > 0:
                    rows_of_selected_param_values.append(row)
        else:
            # weight
            weights = []
            for row in rows_of_weights_and_parameter_values:
                weights.append(row[weight_col])
            del(weights[0]) # header

            # sample rows
            rng = np.random.RandomState(seed=seed)
            sampled_indices = rng.choice(a=range(len(weights)), size=n, p=weights)

            # build sampled rows
            for i in sampled_indices:
                rows_of_selected_param_values.append(rows_of_weights_and_parameter_values[i+1])

        IO.write_csv(rows=rows_of_selected_param_values, file_name=csvfile_resampled_params)

        self.dictOfParamValues = IO.read_csv_cols_to_dictionary(file_name=csvfile_resampled_params,
                                                                if_convert_float=True)
        # check if parameters have values
        for key, values in self.dictOfParamValues.items():
            if len(values) == 0:
                raise ValueError('Parameter values are not provided in '+csvfile_resampled_params+'.')

        for name in columns_to_be_deleted:
            del (self.dictOfParamValues[name])

    def get_mean_interval(self, parameter_name, deci=0, form=None):

        # print the ratio estimate and credible interval
        sum_stat = Stat.SummaryStat('', self.dictOfParamValues[parameter_name])

        if form is None:
            return sum_stat.get_mean(), sum_stat.get_PI(alpha=0.05)
        else:
            return sum_stat.get_formatted_mean_and_interval(
                interval_type='p', alpha=0.05, deci=deci, form=form)

    def plot_histogram(self, parameter_name, title, x_label=None, y_label=None, x_range=None, folder=''):
        """ creates a histogram of one parameter """

        Fig.plot_histogram(
            data=self.dictOfParamValues[parameter_name],
            title=title, x_label=x_label, y_label=y_label,
            x_range=x_range, figure_size=HISTOGRAM_FIG_SIZE, file_name=folder+'/'+title
        )

    def plot_histograms(self, par_names=None, csv_file_name_prior=None, posterior_fig_loc='figures_national'):
        """ creates histograms of parameters specified by ids
        :param par_names: (list) of parameter names to plot
        :param csv_file_name_prior: (string) filename where parameter prior ranges are located
        :param posterior_fig_loc: (string) location where posterior figures_national should be located
        """

        raise ValueError('Needs to be debugged.')

        # clean the directory
        IO.delete_files('.png', posterior_fig_loc)

        # read prior distributions
        dict_of_priors = None
        if csv_file_name_prior is not None:
            dict_of_priors = self.get_dict_of_priors(prior_info_csv_file=csv_file_name_prior)

        if par_names is None:
            par_names = self.get_all_parameter_names()

        # for all parameters, read sampled parameter values and create the histogram
        for par_name in par_names:

            # get values for this parameter
            par_values = self.dictOfParamValues[par_name]

            # get info of this parameter
            title, multiplier, x_range = self.get_title_multiplier_x_range_decimal_format(
                par_name=par_name, dict_of_priors=dict_of_priors)

            # adjust parameter values
            par_values = [v*multiplier for v in par_values]

            # find the filename the histogram should be saved as
            file_name = posterior_fig_loc + '/Par-' + par_name + ' ' + F.proper_file_name(par_name)

            # plot histogram
            Fig.plot_histogram(
                data=par_values,
                title=title.replace('!', '\n'),
                x_range=x_range,
                figure_size=HISTOGRAM_FIG_SIZE,
                file_name=file_name
            )

    def plot_pairwise(self, par_names=None, prior_info_csv_file=None, fig_filename='pairwise_correlation.png',
                      figure_size=(10, 10)):
        """ creates pairwise correlation between parameters specified by ids
        :param par_names: (list) names of parameter to display
        :param prior_info_csv_file: (string) filename where parameter prior ranges are located
            (Note: '!' will be replaced with '\n')
        :param fig_filename: (string) filename to save the figure as
        :param figure_size: (tuple) figure size
        """

        # read information about prior distributions
        dict_of_priors = None
        if prior_info_csv_file is not None:
            dict_of_priors = self.get_dict_of_priors(prior_info_csv_file=prior_info_csv_file)

        # if parameter names are not specified, include all parameters
        if par_names is None:
            par_names = self.get_all_parameter_names()

        # find the info of parameters to include in the analysis
        info_of_param_info_to_include = []

        # for each parameter, read sampled parameter values and create the histogram
        for par_name in par_names:

            # get parameter values
            par_values = self.dictOfParamValues[par_name]

            # get info
            title, multiplier, x_range, decimal, form = self.get_title_multiplier_x_range_decimal_format(
                par_name=par_name, dict_of_priors=dict_of_priors)

            # adjust parameter values
            par_values = [v*multiplier for v in par_values]

            # append the info for this parameter
            info_of_param_info_to_include.append(
                ParamInfo(name=par_name,
                          label=title.replace('!', '\n'),
                          values=par_values, value_range=x_range))

        # plot pairwise
        # set default properties of the figure
        plt.rc('font', size=6)  # fontsize of texts
        plt.rc('axes', titlesize=6)  # fontsize of the figure title
        plt.rc('axes', titleweight='semibold')  # fontweight of the figure title

        # plot each panel
        n = len(info_of_param_info_to_include)

        if n == 0:
            raise ValueError('Values of parameters are not provided. '
                             'Make sure the calibration algorithm exports the parameter values.')

        f, axarr = plt.subplots(nrows=n, ncols=n, figsize=figure_size)

        for i in range(n):
            for j in range(n):

                # get the current axis
                ax = axarr[i, j]

                # plot histogram for diagonal elements
                if i == j:
                    # plot histogram
                    Fig.add_histogram_to_ax(
                        ax=ax,
                        data=info_of_param_info_to_include[i].values,
                        x_range=info_of_param_info_to_include[i].range,
                    )
                    # ax.set_yticklabels([])
                    # ax.set_yticks([])

                else:
                    ax.scatter(info_of_param_info_to_include[j].values,
                               info_of_param_info_to_include[i].values,
                               alpha=0.5, s=2)
                    ax.set_xlim(info_of_param_info_to_include[j].range)
                    ax.set_ylim(info_of_param_info_to_include[i].range)
                    # correlation line
                    b, m = polyfit(info_of_param_info_to_include[j].values,
                                   info_of_param_info_to_include[i].values, 1)
                    ax.plot(info_of_param_info_to_include[j].values,
                            b + m * info_of_param_info_to_include[j].values, '-', c='black')
                    corr, p = pearsonr(info_of_param_info_to_include[j].values,
                                       info_of_param_info_to_include[i].values)
                    ax.text(0.95, 0.95, '{0:.2f}'.format(corr), transform=ax.transAxes, fontsize=6,
                            va='top', ha='right')

                if j == 0:
                    ax.set_ylabel(info_of_param_info_to_include[i].label)
                if i == n-1:
                    ax.set_xlabel(info_of_param_info_to_include[j].label)

        f.align_ylabels(axarr[:, 0])
        f.tight_layout()
        output_figure(plt=f, file_name=fig_filename, dpi=300)

    def export_means_and_intervals(self,
                                   poster_file='ParameterEstimates.csv',
                                   significance_level=0.05, sig_digits=3,
                                   param_names=None, prior_info_csv_file=None):
        """ calculate the mean and credible intervals of parameters specified by ids
        :param poster_file: csv file where the posterior ranges should be stored
        :param significance_level:
        :param sig_digits: number of significant digits
        :param param_names: (list) of parameter names
        :param prior_info_csv_file: (string) filename where parameter prior ranges are located
        :return:
        """

        results = self.get_means_and_intervals(significance_level=significance_level,
                                               sig_digits=sig_digits,
                                               param_names=param_names,
                                               prior_info_csv_file=prior_info_csv_file)

        # write parameter estimates and credible intervals
        IO.write_csv(rows=results, file_name=poster_file)

    def print_means_and_intervals(self,
                                  significance_level=0.05, sig_digits=3,
                                  param_names=None, prior_info_csv_file=None):
        """ calculate the mean and credible intervals of parameters specified by ids
        :param significance_level:
        :param sig_digits: number of significant digits
        :param param_names: (list) of parameter names
        :param prior_info_csv_file: (string) filename where parameter prior ranges are located
        :return:
        """

        results = self.get_means_and_intervals(significance_level=significance_level,
                                               sig_digits=sig_digits, param_names=param_names,
                                               prior_info_csv_file=prior_info_csv_file)
        for r in results:
            print(r)

    def get_means_and_intervals(self, significance_level=0.05, sig_digits=3,
                                param_names=None, prior_info_csv_file=None):
        """
        :param significance_level: (float) significant level to calculate confidence or credible intervals
        :param sig_digits: (int) number of significant digits
        :param param_names: (list) of parameter names
        :param prior_info_csv_file: (string)
        :return:
        """

        results = []  # list of parameter estimates and credible intervals

        # read information about prior distributions
        dict_of_priors = None
        if prior_info_csv_file is not None:
            dict_of_priors = self.get_dict_of_priors(prior_info_csv_file=prior_info_csv_file)

        # if parameter names are not specified, include all parameters
        if param_names is None:
            param_names = self.get_all_parameter_names()

        # for each parameter get mean and  intervals
        for par_name in param_names:

            # get values for this parameter
            par_values = self.dictOfParamValues[par_name]

            # get info of this parameter
            title, multiplier, x_range, decimal, form = self.get_title_multiplier_x_range_decimal_format(
                par_name=par_name, dict_of_priors=dict_of_priors)

            # summary statistics
            sum_stat = Stat.SummaryStat(name=par_name, data=par_values)
            mean_PI_text = sum_stat.get_formatted_mean_and_interval(
                interval_type='p', alpha=significance_level, deci=decimal, sig_digits=sig_digits,
                form=form, multiplier=multiplier)

            results.append(
                [par_name, mean_PI_text])

        return results

    def __calculate_ratio_obss(self, numerator_par_name, denominator_par_names):

        # if only one parameter is in the denominator
        if type(denominator_par_names) is not list:
            denominator_par_names = [denominator_par_names]

        # calculate sum of parameters in the denominator
        sum_denom = copy.deepcopy(self.dictOfParamValues[denominator_par_names[0]])
        for i in range(1, len(denominator_par_names)):
            sum_denom += self.dictOfParamValues[denominator_par_names[i]]

        # calculate realizations for ratio
        return self.dictOfParamValues[numerator_par_name] / sum_denom

    @staticmethod
    def _if_include(par_name, names=None):
        # check if this parameter should be included
        if_include = False
        if names is None:
            if_include = True
        elif par_name in names:
            if_include = True

        return if_include

    def get_ratio_mean_interval(self, numerator_par_name, denominator_par_names, deci=0, form=None):

        # print the ratio estimate and credible interval
        sum_stat = Stat.SummaryStat('', self.__calculate_ratio_obss(numerator_par_name, denominator_par_names))

        if form is None:
            return sum_stat.get_mean(), sum_stat.get_PI(alpha=0.05)
        else:
            return sum_stat.get_formatted_mean_and_interval(
                interval_type='p', alpha=0.05, deci=deci, form=form)

    def plot_ratio_hist(self, numerator_par_name, denominator_par_names,
                        title, x_label=None, x_range=None, output_fig_loc='figures_national'):

        ratio_obss = self.__calculate_ratio_obss(numerator_par_name, denominator_par_names)

        file_name = output_fig_loc + '/Ratio-' + title

        # create the histogram of ratio
        Fig.plot_histogram(
            data=ratio_obss,
            title=title,
            x_label=x_label,
            x_range=x_range,
            figure_size=HISTOGRAM_FIG_SIZE,
            file_name=file_name)

    @staticmethod
    def get_dict_of_priors(prior_info_csv_file):

        dict_priors = {}
        rows_priors = IO.read_csv_rows(
            file_name=prior_info_csv_file,
            if_ignore_first_row=True,
            delimiter=',',
            if_convert_float=True
        )
        for row in rows_priors:
            dict_priors[row[ColumnsPriorDistCSV.NAME.value]] = row

        return dict_priors

    @staticmethod
    def get_title_multiplier_x_range_decimal_format(par_name, dict_of_priors):

        title = par_name
        x_range = None
        multiplier = 1
        decimal = None
        form = None

        if dict_of_priors is not None:
            if par_name in dict_of_priors:

                prior_info = dict_of_priors[par_name]

                if prior_info[ColumnsPriorDistCSV.LB.value] in ('', None) or \
                        prior_info[ColumnsPriorDistCSV.UB.value] in ('', None):
                    x_range = None
                else:
                    try:
                        x_range = [float(prior_info[ColumnsPriorDistCSV.LB.value]),
                                   float(prior_info[ColumnsPriorDistCSV.UB.value])]
                    except:
                        raise ValueError(
                            'Error in reading the prior distribution of parameter {}:'.format(par_name))

                # find title
                if prior_info[ColumnsPriorDistCSV.TITLE.value] in ('', None):
                    title = par_name
                else:
                    title = prior_info[ColumnsPriorDistCSV.TITLE.value]

                # decimal
                decimal = prior_info[ColumnsPriorDistCSV.DECI.value]

                # format
                form = prior_info[ColumnsPriorDistCSV.FORMAT.value]

                # find multiplier
                if prior_info[ColumnsPriorDistCSV.MULTIPLIER.value] in ('', None):
                    multiplier = 1
                else:
                    multiplier = float(prior_info[ColumnsPriorDistCSV.MULTIPLIER.value])

                if x_range is not None:
                    x_range = [x * multiplier for x in x_range]

        return title, multiplier, x_range, decimal, form

    def get_all_parameter_names(self):
        """ :returns the names of all parameters """

        raise ValueError('Needs to be debugged.')

        names = self.dictOfParamValues.keys()

        for to_del in COLUMNS_TO_SKIP:
            del(names[to_del])

        return names



import random

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np

import deampy.plots.plot_support as Fig
import deampy.support.misc_functions as S
import deampy.support.misc_functions as mf
from deampy.plots.plot_support import output_figure


def plot_eff_sample_size(likelihood_weights, if_randomize=True,
                         fig_size=(6, 5), file_name=None,
                         title=None, x_label='Iteration', y_label='Effective Sample Size',
                         x_range=None, y_range=None
                         ):

    # convert the data to np array if needed
    if not type(likelihood_weights) == np.ndarray:
        likelihood_weights = np.array(likelihood_weights)

    # randomize the probabilities if needed
    if if_randomize:
        random.seed(1)
        random.shuffle(likelihood_weights)

    # calculate the effectiveve sample sizes through iterations
    effs = []
    for i in range(len(likelihood_weights)):
        effs.append(S.effective_sample_size(likelihood_weights[:i + 1]))

    # plot
    fig, ax = plt.subplots(figsize=fig_size)
    ax.plot(range(1, len(likelihood_weights) + 1), effs)

    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_xlim(x_range)
    ax.set_ylim(y_range)

    # if f_star is not None:
    #     plt.axhline(y=f_star, linestyle='--', color='black', linewidth=1)
    Fig.output_figure(fig, file_name=file_name)


def plot_mean_stdev_by_np(
        list_of_obss, list_x_ranges, labels, colors,
        y_range_mean=None, y_range_stdev=None,
        y_label_mean=None, y_label_stdev=None,
        x_label=None, # conf_intv_transparency=0.2,
        true_mean=None, true_stdev=None, fig_size=None, file_name=None):
    """
    Plots the cumulative mean and standard deviation of a list of observations against sample size
    :param list_of_obss: list of observations (each observation is a list or np array)
    :param list_x_ranges: list of x ranges (each x range is a list or np array)
    :param labels: list of labels for each time series
    :param colors: list of colors for each time series
    :param y_range_mean: y-axis range for mean plot (list or tuple of form [ymin, ymax])
    :param y_range_stdev: y-axis range for standard deviation plot (list or tuple of form [ymin, ymax])
    :param y_label_mean: y-axis label for mean plot
    :param y_label_stdev: y-axis label for standard deviation plot
    :param x_label: x-axis label for both plots
    :param true_mean: true mean value to plot as a horizontal line (optional)
    :param true_stdev: true standard deviation value to plot as a horizontal line (optional)
    :param fig_size: figure size (tuple of form (width, height))
    :param file_name: file name to save the figure (optional)
    """

    f, axes = plt.subplots(1, 2, figsize=fig_size)
    axes[0].set_title('A)', loc='left', weight='bold')
    axes[1].set_title('B)', loc='left', weight='bold')

    for obs, color, label, x_range in zip(list_of_obss, colors, labels, list_x_ranges):

        cumulative_mean = mf.get_cumulative_mean(obs)
        cumulative_var = np.array(mf.get_cumulative_var(obs), dtype=float)

        axes[0].plot(x_range,
                   cumulative_mean, color=color, linewidth=1,
                   label=label)

        axes[1].plot(x_range,
                   np.sqrt(cumulative_var), color=color, linewidth=1,
                   label=label)
        #
        # if conf_intv_transparency is not None and conf_intv_transparency > 0:
        #     # plot 95% confidence interval for mean
        #     ci_lower, ci_upper = mf.get_confidence_interval_of_mean(obs, conf_level=0.95)
        #     axes[0].fill_between(x_range,
        #                          ci_lower,
        #                          ci_upper,
        #                          color=color,


    if true_mean is not None:
        axes[0].axhline(y=true_mean, color='k', linestyle='--', linewidth=1, label='True Value')

    axes[0].set_xlabel(x_label)
    axes[0].set_ylabel(y_label_mean)
    axes[0].set_ylim(y_range_mean)
    axes[0].legend(fontsize=8)
    # Format x-axis ticks with comma as thousands separator
    axes[0].xaxis.set_major_formatter(ticker.FuncFormatter(lambda x, pos: f"{int(x):,}"))

    axes[1].set_xlabel(x_label)
    axes[1].set_ylabel(y_label_stdev)
    axes[1].set_ylim(y_range_stdev)

    if true_stdev is not None:
        axes[1].axhline(y=true_stdev,
                        color='k', linestyle='--', linewidth=1, label='True Variance')
    # Format x-axis ticks with comma as thousands separator
    axes[1].xaxis.set_major_formatter(ticker.FuncFormatter(lambda x, pos: f"{int(x):,}"))

    output_figure(plt=f, file_name=file_name)


import math


def format_number(number, deci=None, sig_digits=None, format=None):
    """
    :param number: number to be formatted
    :param deci: number of decimal places
    :param sig_digits: number of significant digits
    :param format: additional formatting instruction. 
        Use ',' to format as number, '%' to format as percentage, and '$' to format as currency
    :returns: the text of number with the specified format
    """

    if number is None:
        return ''

    # deci should be an integer
    deci, sig_digits = _get_deci_dig_digit(deci=deci, sig_digits=sig_digits)

    if format not in (None, ',', '%', '$', ''):
        raise ValueError('Invalid value for format.')

    if deci is not None:
        if format is None or format == '' or format is math.nan:
            return '{:.{prec}f}'.format(number, prec=deci)
        elif format == ',':
            return '{:,.{prec}f}'.format(number, prec=deci)
        elif format == '$':
            return '${:,.{prec}f}'.format(number, prec=deci)
        elif format == '%':
            return '{:.{prec}f}%'.format(100*number, prec=deci)

    elif sig_digits is not None:
        if format is None or format == '' or format is math.nan:
            return '{:.{prec}g}'.format(number, prec=sig_digits)
        elif format == ',':
            return '{:,.{prec}g}'.format(number, prec=sig_digits)
        elif format == '$':
            return '${:,.{prec}g}'.format(number, prec=sig_digits)
        elif format == '%':
            return '{:.{prec}g}%'.format(100*number, prec=sig_digits)
    else:
        return str(number)


def format_interval(interval, deci=None, sig_digits=None, format=None):
    """
    :param interval: list of form [low up]
    :param deci: number of decimal places
    :param sig_digits: number of significant digits
    :param format: additional formatting instruction. 
        Use ',' to format as number, '%' to format as percentage, and '$' to format as currency
    :returns a text of form '(low, up)' where the numbers have the specified format """

    if interval is None:
        return ''

    deci, sig_digits = _get_deci_dig_digit(deci=deci, sig_digits=sig_digits)

    if deci is not None:
        if format is None or format == '':
            return '({low:.{prec}f}, {up:.{prec}f})' \
                .format(low=interval[0], up=interval[1], prec=deci)
        elif format == ',':
            return '({low:,.{prec}f}, {up:,.{prec}f})' \
                .format(low=interval[0], up=interval[1], prec=deci)
        elif format == '$':
            return '(${low:,.{prec}f}, ${up:,.{prec}f})' \
                .format(low=interval[0], up=interval[1], prec=deci)
        elif format == '%':
            return '({low:.{prec}f}%, {up:.{prec}f}%)' \
                .format(low=interval[0]*100, up=interval[1]*100, prec=deci)

    elif sig_digits is not None:
        if format is None or format == '':
            return '({low:.{prec}g}, {up:.{prec}g})' \
                .format(low=interval[0], up=interval[1], prec=sig_digits)
        elif format == ',':
            return '({low:,.{prec}g}, {up:,.{prec}g})' \
                .format(low=interval[0], up=interval[1], prec=sig_digits)
        elif format == '$':
            return '(${low:,.{prec}g}, ${up:,.{prec}g})' \
                .format(low=interval[0], up=interval[1], prec=sig_digits)
        elif format == '%':
            return '({low:.{prec}g}%, {up:.{prec}g}%)' \
                .format(low=interval[0]*100, up=interval[1]*100, prec=sig_digits)
    else:
        inside = ', '.join(str(f) for f in interval)
        return '[' + inside + ']'


def format_estimate_interval(estimate, interval, deci=None, sig_digits=None, format=None):
    """
    :param estimate: the estimate
    :param interval: list of form [low up]
    :param deci: number of decimal places
    :param sig_digits: number of significant digits
    :param format: additional formatting instruction. 
        Use ',' to format as number, '%' to format as percentage, and '$' to format as currency
    :return: text in the form 'estimate (l, u)' with the specified decimal places
    """
    return format_number(number=estimate, deci=deci, sig_digits=sig_digits, format=format) \
           + ' ' + format_interval(interval=interval, deci=deci, sig_digits=sig_digits, format=format)


def _get_deci_dig_digit(deci, sig_digits):
    if deci is not None:
        try:
            deci = int(deci)
        except ValueError:
            raise ValueError('deci should be integer or float.')
        except TypeError:
            raise ValueError('deci should be integer or float.')
    elif sig_digits is not None:
        try:
            sig_digits = int(sig_digits)
        except ValueError:
            raise ValueError('sig_digits should be integer or float.')
        except TypeError:
            raise ValueError('sig_digits should be integer or float.')
    else:
        return None, None

    return deci, sig_digits



import csv
import os

import numpy as numpy


class TextFile:
    """
    class to create a text file to add text to
    """

    def __init__(self, filename):
        """
        :param filename: (string) filename to save the text file as
        """

        # make the directory if does not exist
        make_directory(filename)

        # open file
        self.file = open(filename, 'w')

    def write(self, text):
        """
        :param text: (string) text to add to the text file
        """

        self.file.write(text)

    def close(self):
        """ closes the text file """

        self.file.close()


def make_directory(filename):
    """
    creates a directory to save the file for which the filename is provided
    :param filename: (string) name of the file for which the directory should be created for
    """

    # get directory
    directory_path = os.path.dirname(filename)

    # create the directory if does not exist
    if directory_path != '':
        if not os.path.exists(directory_path):
            os.makedirs(directory_path)


def write_csv(rows, file_name='csvfile.csv', delimiter=',', directory='', delete_existing_files=False):
    """ write a list to a csv file
    :param rows: list of lists to be imported to the csv file
    :param file_name: the file name to be given to the csv file
    :param delimiter: to separate by comma, use ',' and by tab, use '\t'
    :param directory: directory (relative to the current root) where the files should be located
            for example use 'Example' to create and save the csv file under the folder Example
    :param delete_existing_files: set to True to delete the existing trace files in the specified directory
    """

    if file_name[0] == '/':
        file_name = file_name[1:]
    if len(directory) > 0 and directory[0] == '/':
        directory = directory[1:]

    # create a new file
    file_name = os.path.join(directory, file_name)

    # delete existing files
    if delete_existing_files:
        delete_files(extension='.csv',
                     path=os.getcwd() + '/' + directory
                     )

    # create the directory if does not exist
    make_directory(filename=file_name)

    with open(file_name, "w", newline='') as file:
        csv_file = csv.writer(file, delimiter=delimiter)

        for row in rows:
            if not (isinstance(row, list) or isinstance(row, numpy.ndarray)):
                csv_file.writerow([row])
            else:
                csv_file.writerow(row)

        file.close()


def write_columns_to_csv(cols, file_name='csvfile.csv', delimiter=',', directory='', delete_existing_files=False):
    """ write a list of columns to a csv file
        :param cols: list of columns to be imported to the csv file
        :param file_name: the file name to be given to the csv file
        :param delimiter: to separate by comma, use ',' and by tab, use '\t'
        :param directory: directory (relative to the current root) where the files should be located
                for example use 'Example' to create and save the csv file under the folder Example
        :param delete_existing_files: set to True to delete the existing trace files in the specified directory
        """
    write_csv(
        rows=_cols_to_rows(cols=cols),
        file_name=file_name, delimiter=delimiter, directory=directory, delete_existing_files=delete_existing_files
    )


def write_dictionary_to_csv(dictionary, file_name='csvfile.csv', directory=''):
    """
    :param dictionary: a dictionary of lists with keys to be csv headers
    :param file_name: filename to save the dictionary as
    """

    cols = []
    for key, values in dictionary.items():
        col = [key]
        col.extend(values)
        cols.append(col)

    write_columns_to_csv(cols=cols, file_name=file_name, directory=directory)


def read_csv_rows(file_name, if_ignore_first_row, delimiter=',', if_convert_float=False):
    """ reads the rows of a csv file
    :param file_name: the csv file name
    :param if_ignore_first_row: set true to ignore the first row
    :param delimiter: to separate by comma, use ',' and by tab, use '\t'
    :param if_convert_float: set true to convert row values to numbers (otherwise, the values are stored as string)
    :returns a list containing the rows of the csv file
    """
    with open(file_name, "r") as file:
        csv_file = csv.reader(file, delimiter=delimiter)

        rows = _csv_file_to_rows(csv_file=csv_file,
                                 if_del_first_row=if_ignore_first_row)

        # convert column values to float if needed
        if if_convert_float:
            for i in range(0, len(rows)):
                rows[i] = _convert_to_float(rows[i])
        file.close()

        return rows


def read_csv_cols(file_name, if_ignore_first_row, n_cols=None, delimiter=',', if_convert_float=False):
    """ reads the columns of a csv file
    :param file_name: the csv file name
    :param n_cols: number of columns in the csv file
    :param if_ignore_first_row: set True to ignore the first row
    :param delimiter: to separate by comma, use ',' and by tab, use '\t'
    :param if_convert_float: set True to convert column values to numbers
    :returns a list containing the columns of the csv file
    """
    with open(file_name, "r", encoding='utf-8', errors='ignore') as file:
        csv_file = csv.reader(file, delimiter=delimiter)

        rows = _csv_file_to_rows(csv_file=csv_file,
                                 if_del_first_row=if_ignore_first_row)

        if n_cols is None:
            n_cols = len(rows[0])

        cols = _rows_to_cols(rows=rows, n_cols=n_cols)

        # convert column values to float if needed
        if if_convert_float:
            for j in range(0, n_cols):
                cols[j] = _convert_to_float(cols[j])

        file.close()

        return cols


def read_csv_cols_to_dictionary(file_name, delimiter=',', if_convert_float=False):

    dict_of_columns = dict()  # dictionary of columns
    csv_file = open(file_name, "r", encoding='utf-8', errors='ignore')
    col_headers = next(csv.reader(csv_file, delimiter=delimiter))
    n_cols = len(col_headers)
    cols = read_csv_cols(
        file_name, 
        n_cols=n_cols,
        if_ignore_first_row=True,
        delimiter=delimiter,
        if_convert_float=if_convert_float)

    csv_file.close()

    # add columns to the dictionary
    for j, col in enumerate(cols):
        if col_headers[j] in dict_of_columns:
            raise ValueError("Key '{}' already exists in the dictionary of columns.".format(col_headers[j]))
        else:
            dict_of_columns[col_headers[j]] = col

    return dict_of_columns


def delete_file(file_name):
    os.remove(file_name)


def delete_files(extension='.txt', path='..'):
    """ delete every files with the specified extension inside the directory
    :param extension: (string) extension of the files to be removed
    :param path: (string) path (relative to the current root) where the files are located
    (the folder should already exist) use os.getcwd() to get the working directory
    """

    try:
        for f in os.listdir(path):
            if f.endswith(extension):
                os.remove(os.path.join(path, f))
    except FileNotFoundError:
        pass


def _csv_file_to_rows(csv_file, if_del_first_row):

    # read rows
    rows = []
    for i, row in enumerate(csv_file):
        rows.append(row)

    # delete the first row if needed
    if if_del_first_row:
        del rows[0]

    return rows


def _rows_to_cols(rows, n_cols):

    # initialize the list to store column values
    cols = []
    for j in range(0, n_cols):
        cols.append([])

    # read columns
    for row in rows:
        if len(row) != n_cols:
            raise ValueError('All rows should have the same length.')

        for j in range(0, n_cols):
            cols[j].append(row[j])

    return cols


def _cols_to_rows(cols):

    # find the size of the largest column
    size_of_largest_column = 0
    for col in cols:
        size = len(col)
        if size > size_of_largest_column:
            size_of_largest_column = size

    # add None values to the columns smaller than the largest column
    for col in cols:
        if size_of_largest_column > len(col):
            col.extend([None]*(size_of_largest_column-len(col)))

    # initialize rows
    rows = []
    for i in range(0, size_of_largest_column):
        rows.append([])

    # populate rows
    for col in cols:
        for i, val in enumerate(col):
            rows[i].append(val)

    return rows


def _convert_to_float(list_of_objs):

    try:
        results = numpy.array(list_of_objs).astype(float)
    except:
        results = []
        for i in range(len(list_of_objs)):
            try:
                x = float(list_of_objs[i])
            except:
                if list_of_objs[i] in ('N/A', 'None', 'none', ''):
                    x = None
                else:
                    x = list_of_objs[i]
            results.append(x)

    return results



import os
import string

from deampy.support.misc_functions import *


def output_figure(plt, file_name=None, dpi=300, tight_layout=True, bbox_inches='tight'):
    """
    :param plt: reference to the plot
    :param file_name: filename to save this figure as (e.g. 'figure.png') (if None, the figure will be displayed)
    :param dpi: dpi of the figure
    """

    if tight_layout:
        plt.tight_layout()

    # output
    if file_name is None or file_name == '':
        plt.show()
    else:
        if file_name[0] == '/':
            file_name = file_name[1:]
        # get directory
        directory_path = os.path.dirname(file_name)

        # create the directory if does not exist
        if directory_path != '':
            if not os.path.exists(directory_path):
                os.makedirs(directory_path)

        try:
            plt.savefig(proper_file_name(file_name), dpi=dpi, bbox_inches=bbox_inches)
        except Exception as e:
            raise ValueError("Error in saving figure '{}'. "
                             "Ensure that the filename is valid. Error Message: {}".format(file_name, str(e)))

        # try:
        #     plt.savefig(proper_file_name(filename), dpi=dpi, bbox_inches=bbox_inches)
        # except ValueError:
        #     raise ValueError("Error in saving figure '{}'. "
        #                      "Ensure that the filename is valid and "
        #                      "that the folder where the figure should be saved exists.".format(filename))


def calculate_ticks(interval, delta):
    values = []
    v = interval[0]
    while v <= interval[1]:
        values.append(v)
        v += delta

    return values


def format_axis_tick_labels(ax, axis='x', format_deci=None):

    if axis == 'x':
        vals = ax.get_xticks()
        if format_deci[0] is None or format_deci[0] == '':
            ax.set_xticklabels(['{:.{prec}f}'.format(x, prec=format_deci[1]) for x in vals])
        elif format_deci[0] == ',':
            ax.set_xticklabels(['{:,.{prec}f}'.format(x, prec=format_deci[1]) for x in vals])
        elif format_deci[0] == '$':
            ax.set_xticklabels(['${:,.{prec}f}'.format(x, prec=format_deci[1]) for x in vals])
        elif format_deci[0] == '%':
            ax.set_xticklabels(['{:,.{prec}%}'.format(x, prec=format_deci[1]) for x in vals])

    elif axis == 'y':
        vals = ax.get_yticks()
        if format_deci[0] is None or format_deci[0] == '':
            ax.set_yticklabels(['{:.{prec}f}'.format(x, prec=format_deci[1]) for x in vals])
        elif format_deci[0] == ',':
            ax.set_yticklabels(['{:,.{prec}f}'.format(x, prec=format_deci[1]) for x in vals])
        elif format_deci[0] == '$':
            ax.set_yticklabels(['${:,.{prec}f}'.format(x, prec=format_deci[1]) for x in vals])
        elif format_deci[0] == '%':
            ax.set_yticklabels(['{:,.{prec}%}'.format(x, prec=format_deci[1]) for x in vals])
    else:
        raise ValueError("axis must be either 'x' or 'y'.")


# def format_x_axis(ax, min_x, max_x, delta_x, buffer=0, form=None, deci=None):
#
#     # get x ticks
#     xs = ax.get_xticks()
#
#     if min_x is None:
#         min_x = xs[0]
#     if max_x is None:
#         max_x = xs[-1]
#     if delta_x is not None:
#         xs = calculate_ticks(min_x, max_x, delta_x)
#
#     # format x-axis
#     ax.set_xticks(xs)
#     if deci is not None:
#         ax.set_xticklabels([F.format_number(x, deci=deci, format=form) for x in xs])
#
#     ax.set_xlim([min_x-buffer, max_x+buffer])


def add_labels_to_panels(axarr, add_to_title=True, x_coord=-0.2, y_coord=1.1, font_size=8):
    """
    adds A), B), etc. labels to panels
    :param axarr: (array) of panels
    :param add_to_title: (bool) add the label to the title (left aligned)
    :param x_coord: (float) increase to move labels right
    :param y_coord: (float) increase to move labels up
    :param font_size: (float) font size of labels
    """
    axs = axarr.flat
    for n, ax in enumerate(axs):
        if add_to_title:
            ax.set_title(
                string.ascii_uppercase[n] + ') ',
                loc='left',
                weight='bold')
        else:
            ax.text(x_coord, y_coord,
                string.ascii_uppercase[n] + ')',
                transform=ax.transAxes,
                size=font_size,
                weight='bold')



import matplotlib.pyplot as plt
import numpy as np

from deampy.plots.plot_support import output_figure, format_axis_tick_labels


def add_histogram_to_ax(ax, data, title=None, label=None, color=None, bin_width=None,
                        x_label=None, y_label=None, x_range=None, y_range=None,
                        transparency=0.75, format_deci=None,
                        linewidth=0.5, x_delta=None):
    """
    :param ax: (axis)
    :param data: (list) observations
    :param title: (string) title of the figure
    :param label: (string) label of this histogram used in the figure legend
    :param x_label: (string) x-axis label
    :param y_label: (string) y-axis label
    :param color: (string) color
    :param bin_width: (integer) bin width
    :param x_range: (list with 2 elements) minimum and maximum of x-axis
    :param y_range: (list with 2 elements) minimum and maximum of y-axis
    :param transparency: (float) between 0 and 1 for the transparency of histogram bins
    :param format_deci: [a, b] where a could be ',', '$', or '%' and b is the decimal point
    :param linewidth: (double) width of histogram lines
    :param x_delta: (double) distance between x_axis ticks and labels
    :return:
    """

    ax.hist(data,
            bins=find_bins(data, x_range, bin_width),
            color=color,
            edgecolor='black',
            linewidth=linewidth,
            alpha=transparency,
            label=label)
    ax.set_xlim(x_range)
    ax.set_title(title)
    # ax.yaxis.set_visible(not remove_y_labels)
    if y_label is None:
        ax.set_yticklabels([])
        ax.set_yticks([])

    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if x_delta is not None and x_range is not None:
        vals_x = []
        x = x_range[0]
        while x <= x_range[1]:
            vals_x.append(x)
            x += x_delta
        ax.set_xticks(vals_x)

    ax.set_xlim(x_range)
    ax.set_ylim(y_range)

    if format_deci is not None:
        format_axis_tick_labels(ax=ax, axis='x', format_deci=format_deci)
        # vals = ax.get_xticks()
        # if format_deci[0] is None or format_deci[0] == '':
        #     ax.set_xticklabels(['{:.{prec}f}'.format(x, prec=format_deci[1]) for x in vals])
        # elif format_deci[0] == ',':
        #     ax.set_xticklabels(['{:,.{prec}f}'.format(x, prec=format_deci[1]) for x in vals])
        # elif format_deci[0] == '$':
        #     ax.set_xticklabels(['${:,.{prec}f}'.format(x, prec=format_deci[1]) for x in vals])
        # elif format_deci[0] == '%':
        #     ax.set_xticklabels(['{:,.{prec}%}'.format(x, prec=format_deci[1]) for x in vals])


def plot_histogram(data, title=None,
                   x_label=None, y_label=None, bin_width=None, transparency=0.5,
                   x_range=None, y_range=None, figure_size=None,
                   color=None, legend=None, file_name=None):
    """ plot a histogram
    :param data: (list) observations
    :param title: (string) title of the figure
    :param x_label: (string) x-axis label
    :param y_label: (string) y-axis label
    :param bin_width: bin width
    :param transparency: (float) between 0 and 1 for the transparency of histogram bins
    :param x_range: (list with 2 elements) minimum and maximum of x-axis
    :param y_range: (list with 2 elements) minimum and maximum of y-axis
    :param figure_size: (tuple) figure size
    :param color: (string) color
    :param legend: string for the legend
    :param file_name: (string) filename to save the histogram as (e.g. 'fig.png')
    """

    fig, ax = plt.subplots(figsize=figure_size)

    # add histogram
    add_histogram_to_ax(ax=ax,
                        data=data,
                        color=color,
                        bin_width=bin_width,
                        title=title,
                        x_label=x_label,
                        y_label=y_label,
                        x_range=x_range,
                        y_range=y_range,
                        transparency=transparency)

    # add legend if provided
    if legend is not None:
        ax.legend([legend])

    # output figure
    output_figure(fig, file_name)


def add_histograms_to_ax(ax, data_sets, legends=None, legend_fontsize=8, bin_width=None,
                         title=None, x_label=None, y_label=None,
                         x_range=None, y_range=None,
                         color_codes=None, transparency=1):

    # add histograms
    for i, data in enumerate(data_sets):
        color = None
        if color_codes is not None:
            color = color_codes[i]

        labels = []
        if legends is None:
            labels = [None]*len(data_sets)
        else:
            labels = legends

        add_histogram_to_ax(ax=ax,
                            data=data,
                            bin_width=bin_width,
                            x_range=x_range,
                            y_label=y_label,
                            color=color,
                            transparency=transparency,
                            label=labels[i])

    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_xlim(x_range)
    ax.set_ylim(y_range)
    if legends is not None:
        ax.legend(fontsize=legend_fontsize)


def plot_histograms(data_sets, legends, bin_width=None,
                    title=None, x_label=None, y_label=None,
                    x_range=None, y_range=None, figure_size=None,
                    color_codes=None, transparency=1, file_name=None):
    """
    plots multiple histograms on a single figure
    :param data_sets: (list of lists) observations
    :param legends: (list) string for the legend
    :param bin_width: bin width
    :param title: (string) title of the figure
    :param x_label: (string) x-axis label
    :param y_label: (string) y-axis label
    :param x_range: (list with 2 elements) minimum and maximum of x-axis
    :param y_range: (list with 2 elements) minimum and maximum of y-axis
    :param figure_size: (tuple) figure size
    :param color_codes: (list) of colors
    :param transparency: (float) 0.0 transparent through 1.0 opaque
    :param file_name: (string) filename to save the histogram as (e.g. 'fig.png')
    """

    fig, ax = plt.subplots(figsize=figure_size)
    add_histograms_to_ax(ax=ax, data_sets=data_sets,
                         legends=legends, bin_width=bin_width,
                         title=title, x_label=x_label, y_label=y_label,
                         x_range=x_range, y_range=y_range,
                         color_codes=color_codes, transparency=transparency)

    # output figure
    output_figure(plt, file_name)


def find_bins(data, x_range, bin_width):

    if bin_width is None:
        return 'auto'

    if x_range is not None:
        l = x_range[0]
        u = x_range[1]
    else:
        l = min(data)
        u = max(data) + bin_width
    return np.arange(l, u, bin_width)



import deampy.statistics as Stat


class _SamplePath:

    def __init__(self, name, sim_rep=0, collect_stat=True, warm_up_period=0):
        """
        :param name: name of this sample path
        :param sim_rep: (int) simulation replication of this sample path
        :param collect_stat: set to True to collect statistics on
                            average, max, min, stDev, etc for this sample path
        :param warm_up_period: warm up period (observations before this time will not be used to calculate statistics)
        """
        self.name = name
        self.simRep = sim_rep
        self._period_num = 0
        self._warm_up_period = warm_up_period

        self._times = []  # times at which observations should be recorded
        self._values = []  # value of this sample path over time
        # statistics on this prevalence sample path
        self.ifCollectStat = collect_stat

    def record_increment(self, time, increment):
        raise NotImplemented()

    def record_value(self, time, value):
        raise NotImplemented()

    def close(self, time):
        raise NotImplemented()

    def get_values(self, delete_initial_zeroes=False):
        if delete_initial_zeroes:
            # delete initial zeroes
            if len(self._values) > 0 and self._values[0] == 0:
                i = 0
                while i < len(self._values) and self._values[i] == 0:
                    i += 1
                return self._values[i:]
            else:
                return self._values
        else:
            return self._values

    def get_current_value(self):
        return self._values[-1]


class PrevalenceSamplePath(_SamplePath):

    def __init__(self, name, initial_size=0, sim_rep=0, collect_stat=True, ave_method='step', warm_up_period=0):
        """
        :param name: name of this sample path.
        :param initial_size: value of the sample path at simulation time 0.
        :param sim_rep: (int) simulation replication of this sample path.
        :param collect_stat: set to True to collect statistics
                        on average, max, min, stDev, etc for this sample path.
        :param ave_method: to calculate the area under the curve.
            'step' assumes that changes occurred at the time where observations are recorded,
            'linear' assumes uses trapezoid approach to calculate the area under the curve
        :param warm_up_period: warm up period (observations before this time will not be used to calculate statistics)
        """
        _SamplePath.__init__(self, name=name, sim_rep=sim_rep,
                             collect_stat=collect_stat, warm_up_period=warm_up_period)
        self.currentSize = initial_size  # current size of the sample path
        self._times = [0]
        self._values = [initial_size]
        # statistics on this prevalence sample path
        if collect_stat:
            self.stat = Stat.ContinuousTimeStat(name=name, initial_time=warm_up_period, ave_method=ave_method)

    def record_increment(self, time, increment):
        """
        updates the value of this sample path (e.g. number of people in the system)
        :param time: time of this change
        :param increment: (integer) change (+ or -) in value of this sample path
        """

        if time < self._times[-1]:
            raise ValueError(self.name + ' | Current time cannot be less than the last recorded time.')
        if increment is None:
            raise ValueError(self.name + ' | increment cannot be None.')

        self.currentSize += increment

        # update continuous-time stat
        if self.ifCollectStat:
            self.stat.record(time=time, increment=increment)

        if time == self._times[-1]:
            self._values[-1] = self.currentSize
        else:
            self._times.append(time)
            self._values.append(self.currentSize)

    def record_value(self, time, value):
        """
        updates the value of this sample path (e.g. number of people in the system)
        :param time: time of this change
        :param value: the current value of this sample path
        """

        self.record_increment(time=time, increment=value-self.currentSize)

    def populate(self, times, values):
        """
        populates the sample path with the given times and values
        :param times: (list) times at which the sample path are recorded
        :param values: (list) values of the sample path at the given times
        """

        if len(times) != len(values):
            raise ValueError(self.name + ' | The list of times should be the same size as the list of values.')

        for i in range(len(times)):
            self.record_value(time=times[i], value=values[i])

    def close(self, time):
        self.record_increment(time=time, increment=0)

    def get_times(self):
        return self._times

    def get_mean(self):
        if self.ifCollectStat:
            return self.stat.get_mean()
        else:
            raise ValueError(self.name + ' | Statistics were not collected for this sample path. '
                                         'Set collect_stat = True when initializing the PrevalenceSamplePath class.')


class IncidenceSamplePath(_SamplePath):
    def __init__(self, name, delta_t, sim_rep=0, collect_stat=True, warm_up_period=0):
        """
        :param name: name of this sample path
        :param delta_t: length of equally-spaced observation periods
        :param sim_rep: (int) simulation replication of this sample path
        :param collect_stat: set to True to collect statistics on average, max, min, stDev, etc for this sample path
        :param warm_up_period: warm up period (observations before this time will not be used to calculate statistics)
        """
        _SamplePath.__init__(self, name=name, sim_rep=sim_rep,
                             collect_stat=collect_stat, warm_up_period=warm_up_period)
        self._deltaT = delta_t
        self._period_num = 0
        self._warm_up_period = warm_up_period
        self._period_nums = []  # times represent the observation period number
        self._values = []
        # statistics on this incidence sample path
        if self.ifCollectStat:
            self.stat = Stat.DiscreteTimeStat(name=name)

    def record_increment(self, time, increment):
        """
        updates the value of this sample path (e.g. number of people in the system)
        :param time: time of this change
        :param increment: (integer) change (+ or -) in value of this sample path
        """

        if len(self._period_nums) > 0 and time < (self._period_nums[-1]-1)*self._deltaT:
            raise ValueError(self.name + ' | Current time cannot be less than the last recorded time.')
        if increment is None:
            raise ValueError(self.name + ' | increment cannot be None.')
        if increment < 0:
            raise ValueError(self.name + ' | increment cannot be negative.')

        if time >= self._period_num * self._deltaT:
            self.__fill_unrecorded_periods(time=time)

            self._period_num += 1
            self._values.append(increment)
            self._period_nums.append(self._period_num)

            if self.ifCollectStat and time >= self._warm_up_period:
                self.stat.record(obs=self._values[-1])

        else:
            self._values[-1] += increment

    def record_value(self, time, value):
        """
        updates the value of this sample path (e.g. number of people diagnosed in the past year)
        :param time:
        :param value:
        """

        self.record_increment(time=time, increment=value)

    def populate(self, times, values):
        """
        populates the sample path with the given times and values
        :param times: (list) times at which the sample path are recorded
        :param values: (list) values of the sample path at the given times
        """

        if len(times) != len(values):
            raise ValueError(self.name + ' | The list of times should be the same size as the list of values.')

        for i in range(len(times)):
            self.record_value(time=times[i], value=values[i])

    def close(self, time):
        self.__fill_unrecorded_periods(time=time)

    def get_period_numbers(self):
        return self._period_nums

    def __fill_unrecorded_periods(self, time):
        while time >= (self._period_num + 1) * self._deltaT:
            self._period_num += 1
            self._values.append(0)
            self._period_nums.append(self._period_num)

            if self.ifCollectStat:
                self.stat.record(obs=self._values[-1])


class PrevalencePathBatchUpdate(PrevalenceSamplePath):
    """ a sample path for which observations are recorded at the end of the simulation """

    def __init__(self, name, initial_size, times_of_changes, increments, sim_rep=0):
        """
        :param name: name of this sample path
        :param initial_size: (int) value of the sample path at simulation time 0
        :param times_of_changes: (list) times at which changes occurred
        :param increments: (list) level of changes
        :param sim_rep: (int) simulation replication of this sample path
        """

        if len(times_of_changes) != len(increments):
            raise ValueError('The list of times at which changes occurred should '
                             'be the same size as the list of changes.')

        PrevalenceSamplePath.__init__(self, name=name, initial_size=initial_size, sim_rep=sim_rep)

        # populate the list of [time, value] of recordings
        self._time_and_values = []
        for i in range(len(times_of_changes)):
            self._time_and_values.append([times_of_changes[i], increments[i]])

        # the batch sample path will be converted to real-time sample path
        self._samplePath = PrevalenceSamplePath(name, initial_size, sim_rep)
        self._ifProcessed = False   # set to True when the sample path is built

    def record(self, time, increment):
        pass

    def get_times(self):
        """
        :return: times when changes in the sample path recorded
        """

        # build the sample path if not build yet
        if not self._ifProcessed:
            self.__process()
        return self._samplePath._times

    def get_values(self):
        """
        :return: value of the sample path at times when changes in the sampel path recorded
        """

        # build the sample path if not build yet
        if not self._ifProcessed:
            self.__process()
        return self._samplePath._values

    def __process(self):
        """ will build the sample path when requested """

        # sort this list based on the recorded time
        new_list = sorted(self._time_and_values, key=lambda x: x[0])

        # create the sample path
        for item in new_list:
            self._samplePath.record_increment(item[0], item[1])

        # proceed
        self._ifProcessed = True



from .optimization import learning_exploring_rules, opt_support, stoch_approx
from .plots import econ_eval_plots, histogram, plot_support, pop_pyramids, prob_dists, sample_paths
from .support import econ_eval_support, misc_functions, misc_classes, simulation



import os

from numpy import iinfo, int32

import deampy.in_out_functions as io


class _Trace:
    def __init__(self, if_should_trace, deci):
        """ creates a list to store trace messages
        :param if_should_trace: enter true to trace the simulation
        :param deci: number of decimals to round time values to
        """

        self._on = if_should_trace          # if set to False, the trace messages will not be stored
        self._deci = deci
        self._messages = []                 # list of strings to store trace messages
        self._tOfLastMessage = 0            # time when the last message was recorded

    def _add_message(self, time, message):
        """ adds the entered text to the trace list
        :param time: (float) current simulation time
        :param message: (string) the message to describe what happened at the current time
        """
        if not self._on:
            return

        # if the time has changed since the last message, add an empty message
        if time > self._tOfLastMessage:
            self._messages.append('---')

        # message
        text = "At {t:.{prec}f}: ".format(t=time, prec=self._deci) + message

        # record the message
        self._messages.append(text)

        # update the time of last message
        self._tOfLastMessage = time

    def get_trace(self):
        """
        :return: the list of trace messages
        """
        if not self._on:
            return

        return self._messages

    def print_trace(self, filename, directory='Trace', delete_existing_files=True):
        """ print the trace messages into a text file with the specified filename.
        It creates a sub directory where the python script is located.
        :param filename: filename of the text file where trace message should be exported to
        :param directory: directory (relative to the current root) where the trace files should be located
        :param delete_existing_files: set to True to delete the existing trace files in the directory
        """
        if not self._on:
            return

        # create the directory if does not exist
        if not os.path.exists(directory):
            os.makedirs(directory)

        # delete existing files
        if delete_existing_files:
            io.delete_files(extension='.txt', path=os.getcwd()+'/'+directory)

        # create a new file
        filename = os.path.join(directory, filename)
        # open the file with a write access
        file = open(filename, 'w')
        # write the trace messages
        for message in self._messages:
            file.write('%s\n' % message)
        # close the file
        file.close()


class Trace(_Trace):
    def __init__(self, if_should_trace, deci):
        """ creates a list to store trace messages
        :param if_should_trace: enter true to trace the simulation
        :param deci: number of decimals to round time values to
        """
        _Trace.__init__(self, if_should_trace=if_should_trace, deci=deci)

    def add_message(self, time, message):
        """ adds the entered text to the trace list
        :param time: (float) current simulation time
        :param message: (string) the message to describe what happened at the current time
        """
        self._add_message(time=time, message=message)


class DiscreteEventSimTrace(_Trace):
    # trace for discrete-event simulation models

    def __init__(self, sim_calendar, if_should_trace, deci):
        """ creates a list to store trace messages
        :param sim_calendar: simulation calendar
        :param if_should_trace: enter true to trace the simulation
        :param deci: number of decimals to round time values to
        """

        _Trace.__init__(self, if_should_trace=if_should_trace, deci=deci)
        self._simCalendar = sim_calendar    # simulation calender (to get the current simulation time)

    def add_message(self, message):
        """ adds the entered text to the trace list
        :param message: (string) the message to describe what happened at the current time
        """

        self._add_message(time=self._simCalendar.time, message=message)


class SeedGenerator:

    def __init__(self, seeds=None, weights=None):
        """
        :param seeds: (list) of seeds
        :param weights: (list) of seed weighs
        """

        if seeds is not None and weights is not None:
            if len(seeds) != len(weights):
                raise ValueError('There should be equal number of seeds and weights.')
            if sum(weights) == 0:
                raise ValueError('Weights of all provided seeds are zero. '
                                 'Make sure that at least one seed has a positive weight.')

        self.seeds = seeds
        self.weights = weights
        self.max = iinfo(int32).max

        self.i = -1
        self.posWeights = []

        if seeds is not None and weights is None:
            self.seedsWithPosWeights = seeds
        else:
            self.seedsWithPosWeights = []
            if self.weights is not None:
                for s, w in zip(self.seeds, self.weights):
                    if w > 0:
                        self.seedsWithPosWeights.append(s)
                        self.posWeights.append(w)

    def next_seed(self, rng=None, sample_by_weight=False):
        """
        :param rng: (Random number generator)
        :param sample_by_weight: (bool) True to sample seeds by weight and
                                        False to only used seeds with positive weights
        """

        if self.seeds is not None:
            # if seeds are provided
            if sample_by_weight:
                # if sample seeds according to weights
                return rng.choice(a=self.seeds, size=1, p=self.weights)
            else:
                # if seeds should not be sampled, return the next seed with positive weight
                if self.i + 1 < len(self.seedsWithPosWeights):
                    self.i += 1
                else:
                    self.i = 0
                return self.seedsWithPosWeights[self.i]
        else:
            # if seeds are not provided, return a random integer
            return rng.randint(0, self.max)

    def next_seeds(self, n, rng=None, sample_by_weight=False):

        seeds = []
        for i in range(n):
            seeds.append(self.next_seed(rng=rng, sample_by_weight=sample_by_weight))

        return seeds
